# (c) 2012-2014, Michael DeHaan <michael.dehaan@gmail.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import dataclasses
import errno
import os
import sys
import signal
import tempfile
import threading
import time
import typing as t
import multiprocessing.queues

from ansible import constants as C
from ansible import context
from ansible.errors import AnsibleError, ExitCode, AnsibleCallbackError
from ansible._internal._errors._handler import ErrorHandler
from ansible.executor.play_iterator import PlayIterator
from ansible.executor.stats import AggregateStats
from ansible.executor.task_result import _RawTaskResult, _WireTaskResult
from ansible.inventory.data import InventoryData
from ansible.module_utils.common.text.converters import to_native
from ansible.parsing.dataloader import DataLoader
from ansible.playbook.play_context import PlayContext
from ansible.playbook.task import Task
from ansible.plugins.loader import callback_loader, strategy_loader, module_loader
from ansible.plugins.callback import CallbackBase
from ansible._internal._templating._engine import TemplateEngine
from ansible.vars.hostvars import HostVars
from ansible.vars.manager import VariableManager
from ansible.utils.display import Display
from ansible.utils.lock import lock_decorator
from ansible.utils.multiprocessing import context as multiprocessing_context

if t.TYPE_CHECKING:
    from ansible.executor.process.worker import WorkerProcess

__all__ = ['TaskQueueManager']

STDIN_FILENO = 0
STDOUT_FILENO = 1
STDERR_FILENO = 2

display = Display()

_T = t.TypeVar('_T')


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class CallbackSend:
    method_name: str
    wire_task_result: _WireTaskResult


class DisplaySend:
    def __init__(self, method, *args, **kwargs):
        self.method = method
        self.args = args
        self.kwargs = kwargs


@dataclasses.dataclass
class PromptSend:
    worker_id: int
    prompt: str
    private: bool = True
    seconds: int = None
    interrupt_input: t.Iterable[bytes] = None
    complete_input: t.Iterable[bytes] = None


class FinalQueue(multiprocessing.queues.SimpleQueue):
    def __init__(self, *args, **kwargs):
        kwargs['ctx'] = multiprocessing_context
        super().__init__(*args, **kwargs)

    def send_callback(self, method_name: str, task_result: _RawTaskResult) -> None:
        self.put(CallbackSend(method_name=method_name, wire_task_result=task_result.as_wire_task_result()))

    def send_task_result(self, task_result: _RawTaskResult) -> None:
        self.put(task_result.as_wire_task_result())

    def send_display(self, method, *args, **kwargs):
        self.put(
            DisplaySend(method, *args, **kwargs),
        )

    def send_prompt(self, **kwargs):
        self.put(
            PromptSend(**kwargs),
        )


class AnsibleEndPlay(Exception):
    def __init__(self, result):
        self.result = result


class TaskQueueManager:

    """
    This class handles the multiprocessing requirements of Ansible by
    creating a pool of worker forks, a result handler fork, and a
    manager object with shared datastructures/queues for coordinating
    work between all processes.

    The queue manager is responsible for loading the play strategy plugin,
    which dispatches the Play's tasks to hosts.
    """

    RUN_OK = ExitCode.SUCCESS
    RUN_ERROR = ExitCode.GENERIC_ERROR
    RUN_FAILED_HOSTS = ExitCode.HOST_FAILED
    RUN_UNREACHABLE_HOSTS = ExitCode.HOST_UNREACHABLE
    RUN_FAILED_BREAK_PLAY = 8  # never leaves PlaybookExecutor.run
    RUN_UNKNOWN_ERROR = 255  # never leaves PlaybookExecutor.run, intentionally includes the bit value for 8

    _callback_dispatch_error_handler = ErrorHandler.from_config('_CALLBACK_DISPATCH_ERROR_BEHAVIOR')

    def __init__(
        self,
        inventory: InventoryData,
        variable_manager: VariableManager,
        loader: DataLoader,
        passwords: dict[str, str | None],
        stdout_callback_name: str | None = None,
        run_additional_callbacks: bool = True,
        run_tree: bool = False,
        forks: int | None = None,
    ) -> None:
        self._inventory = inventory
        self._variable_manager = variable_manager
        self._loader = loader
        self._stats = AggregateStats()
        self.passwords = passwords
        self._stdout_callback_name: str | None = stdout_callback_name or C.DEFAULT_STDOUT_CALLBACK
        self._run_additional_callbacks = run_additional_callbacks
        self._run_tree = run_tree
        self._forks = forks or 5

        self._callback_plugins: list[CallbackBase] = []
        self._start_at_done = False

        # make sure any module paths (if specified) are added to the module_loader
        if context.CLIARGS.get('module_path', False):
            for path in context.CLIARGS['module_path']:
                if path:
                    module_loader.add_directory(path)

        # a special flag to help us exit cleanly
        self._terminated = False

        # dictionaries to keep track of failed/unreachable hosts
        self._failed_hosts: dict[str, t.Literal[True]] = dict()
        self._unreachable_hosts: dict[str, t.Literal[True]] = dict()

        try:
            self._final_q = FinalQueue()
        except OSError as e:
            raise AnsibleError("Unable to use multiprocessing, this is normally caused by lack of access to /dev/shm: %s" % to_native(e))

        try:
            # Done in tqm, and not display, because this is only needed for commands that execute tasks
            for fd in (STDIN_FILENO, STDOUT_FILENO, STDERR_FILENO):
                os.set_inheritable(fd, False)
        except Exception as ex:
            display.error_as_warning("failed to set stdio as non inheritable", exception=ex)

        self._callback_lock = threading.Lock()

        # A temporary file (opened pre-fork) used by connection
        # plugins for inter-process locking.
        self._connection_lockfile = tempfile.TemporaryFile()

        self._workers: list[WorkerProcess | None] = []

        # signal handlers to propagate signals to workers
        signal.signal(signal.SIGTERM, self._signal_handler)
        signal.signal(signal.SIGINT, self._signal_handler)

    def _initialize_processes(self, num: int) -> None:
        # mutable update to ensure the reference stays the same
        self._workers[:] = [None] * num

    def _signal_handler(self, signum, frame) -> None:
        """
        terminate all running process groups created as a result of calling
        setsid from within a WorkerProcess.

        Since the children become process leaders, signals will not
        automatically propagate to them.
        """
        signal.signal(signum, signal.SIG_DFL)

        for worker in self._workers:
            if worker is None or not worker.is_alive():
                continue
            if worker.pid:
                try:
                    # notify workers
                    os.kill(worker.pid, signum)
                except OSError as e:
                    if e.errno != errno.ESRCH:
                        signame = signal.strsignal(signum)
                        display.error(f'Unable to send {signame} to child[{worker.pid}]: {e}')

        if signum == signal.SIGINT:
            # Defer to CLI handling
            raise KeyboardInterrupt()

        pid = os.getpid()
        try:
            os.kill(pid, signum)
        except OSError as e:
            signame = signal.strsignal(signum)
            display.error(f'Unable to send {signame} to {pid}: {e}')

    def load_callbacks(self):
        """
        Loads all available callbacks, with the exception of those which
        utilize the CALLBACK_TYPE option. When CALLBACK_TYPE is set to 'stdout',
        only one such callback plugin will be loaded.
        """

        if self._callback_plugins:
            return

        if not self._stdout_callback_name:
            raise AnsibleError("No stdout callback name provided.")

        stdout_callback = callback_loader.get(self._stdout_callback_name)

        if not stdout_callback:
            raise AnsibleError(f"Could not load {self._stdout_callback_name!r} callback plugin.")

        stdout_callback._init_callback_methods()
        stdout_callback.set_options()

        self._callback_plugins.append(stdout_callback)

        # get all configured loadable callbacks (adjacent, builtin)
        plugin_types = {plugin_type.ansible_name: plugin_type for plugin_type in callback_loader.all(class_only=True)}

        # add enabled callbacks that refer to collections, which might not appear in normal listing
        for c in C.CALLBACKS_ENABLED:
            # load all, as collection ones might be using short/redirected names and not a fqcn
            plugin = callback_loader.get(c, class_only=True)

            if plugin:
                # avoids incorrect and dupes possible due to collections
                plugin_types.setdefault(plugin.ansible_name, plugin)
            else:
                display.warning("Skipping callback plugin '%s', unable to load" % c)

        plugin_types.pop(stdout_callback.ansible_name, None)

        # for each callback in the list see if we should add it to 'active callbacks' used in the play
        for callback_plugin in plugin_types.values():
            callback_type = getattr(callback_plugin, 'CALLBACK_TYPE', '')
            callback_needs_enabled = getattr(callback_plugin, 'CALLBACK_NEEDS_ENABLED', getattr(callback_plugin, 'CALLBACK_NEEDS_WHITELIST', False))

            # try to get collection world name first
            cnames = getattr(callback_plugin, '_redirected_names', [])
            if cnames:
                # store the name the plugin was loaded as, as that's what we'll need to compare to the configured callback list later
                callback_name = cnames[0]
            else:
                # fallback to 'old loader name'
                (callback_name, ext) = os.path.splitext(os.path.basename(callback_plugin._original_path))

            display.vvvvv("Attempting to use '%s' callback." % (callback_name))
            if callback_type == 'stdout':
                # we only allow one callback of type 'stdout' to be loaded,
                display.vv("Skipping callback '%s', as we already have a stdout callback." % (callback_name))
                continue
            elif callback_name == 'tree' and self._run_tree:
                # TODO: remove special case for tree, which is an adhoc cli option --tree
                pass
            elif not self._run_additional_callbacks or (callback_needs_enabled and (
                # only run if not adhoc, or adhoc was specifically configured to run + check enabled list
                    C.CALLBACKS_ENABLED is None or callback_name not in C.CALLBACKS_ENABLED)):
                # 2.x plugins shipped with ansible should require enabling, older or non shipped should load automatically
                continue

            try:
                callback_obj = callback_plugin()
                # avoid bad plugin not returning an object, only needed cause we do class_only load and bypass loader checks,
                # really a bug in the plugin itself which we ignore as callback errors are not supposed to be fatal.
                if callback_obj:
                    callback_obj._init_callback_methods()
                    callback_obj.set_options()
                    self._callback_plugins.append(callback_obj)
                else:
                    display.warning("Skipping callback '%s', as it does not create a valid plugin instance." % callback_name)
                    continue
            except Exception as ex:
                display.error_as_warning(f"Failed to load callback plugin {callback_name!r}.", exception=ex)
                continue

    def run(self, play):
        """
        Iterates over the roles/tasks in a play, using the given (or default)
        strategy for queueing tasks. The default is the linear strategy, which
        operates like classic Ansible by keeping all hosts in lock-step with
        a given task (meaning no hosts move on to the next task until all hosts
        are done with the current task).
        """

        self.load_callbacks()

        all_vars = self._variable_manager.get_vars(play=play)
        templar = TemplateEngine(loader=self._loader, variables=all_vars)

        new_play = play.copy()
        new_play.post_validate(templar)
        new_play.handlers = new_play.compile_roles_handlers() + new_play.handlers

        self.hostvars = HostVars(
            inventory=self._inventory,
            variable_manager=self._variable_manager,
            loader=self._loader,
        )

        play_context = PlayContext(new_play, self.passwords, self._connection_lockfile.fileno())

        for callback_plugin in self._callback_plugins:
            callback_plugin.set_play_context(play_context)

        self.send_callback('v2_playbook_on_play_start', new_play)

        # build the iterator
        iterator = PlayIterator(
            inventory=self._inventory,
            play=new_play,
            play_context=play_context,
            variable_manager=self._variable_manager,
            all_vars=all_vars,
            start_at_done=self._start_at_done,
        )

        # adjust to # of workers to configured forks or size of batch, whatever is lower
        self._initialize_processes(min(self._forks, iterator.batch_size))

        # load the specified strategy (or the default linear one)
        strategy = strategy_loader.get(new_play.strategy, self)
        if strategy is None:
            raise AnsibleError("Invalid play strategy specified: %s" % new_play.strategy, obj=play._ds)

        # Because the TQM may survive multiple play runs, we start by marking
        # any hosts as failed in the iterator here which may have been marked
        # as failed in previous runs. Then we clear the internal list of failed
        # hosts so we know what failed this round.
        for host_name in self._failed_hosts.keys():
            host = self._inventory.get_host(host_name)
            iterator.mark_host_failed(host)
        for host_name in self._unreachable_hosts.keys():
            iterator._play._removed_hosts.append(host_name)

        self.clear_failed_hosts()

        # during initialization, the PlayContext will clear the start_at_task
        # field to signal that a matching task was found, so check that here
        # and remember it so we don't try to skip tasks on future plays
        if context.CLIARGS.get('start_at_task') is not None and play_context.start_at_task is None:
            self._start_at_done = True

        # and run the play using the strategy and cleanup on way out
        try:
            play_return = strategy.run(iterator, play_context)
        finally:
            strategy.cleanup()
            self._cleanup_processes()

        # now re-save the hosts that failed from the iterator to our internal list
        for host_name in iterator.get_failed_hosts():
            self._failed_hosts[host_name] = True

        if iterator.end_play:
            raise AnsibleEndPlay(play_return)

        return play_return

    def cleanup(self):
        display.debug("RUNNING CLEANUP")
        self.terminate()
        self._final_q.close()
        self._cleanup_processes()
        # We no longer flush on every write in ``Display.display``
        # just ensure we've flushed during cleanup
        sys.stdout.flush()
        sys.stderr.flush()

    def _cleanup_processes(self):
        if hasattr(self, '_workers'):
            for attempts_remaining in range(C.WORKER_SHUTDOWN_POLL_COUNT - 1, -1, -1):
                if not any(worker_prc and worker_prc.is_alive() for worker_prc in self._workers):
                    break

                if attempts_remaining:
                    time.sleep(C.WORKER_SHUTDOWN_POLL_DELAY)
                else:
                    display.warning('One or more worker processes are still running and will be terminated.')

            for worker_prc in self._workers:
                if worker_prc and worker_prc.is_alive():
                    try:
                        worker_prc.terminate()
                    except AttributeError:
                        pass

    def clear_failed_hosts(self) -> None:
        self._failed_hosts = dict()

    def get_inventory(self) -> InventoryData:
        return self._inventory

    def get_variable_manager(self) -> VariableManager:
        return self._variable_manager

    def get_loader(self) -> DataLoader:
        return self._loader

    def get_workers(self):
        return self._workers[:]

    def terminate(self) -> None:
        self._terminated = True

    def has_dead_workers(self) -> bool:

        # [<WorkerProcess(WorkerProcess-2, stopped[SIGKILL])>,
        # <WorkerProcess(WorkerProcess-2, stopped[SIGTERM])>

        defunct = False
        for x in self._workers:
            if getattr(x, 'exitcode', None):
                defunct = True
        return defunct

    @staticmethod
    def _first_arg_of_type(value_type: t.Type[_T], args: t.Sequence) -> _T | None:
        return next((arg for arg in args if isinstance(arg, value_type)), None)

    @lock_decorator(attr='_callback_lock')
    def send_callback(self, method_name, *args, **kwargs):
        # We always send events to stdout callback first, rest should follow config order
        for callback_plugin in self._callback_plugins:
            # a plugin that set self.disabled to True will not be called
            # see osx_say.py example for such a plugin
            if callback_plugin.disabled:
                continue

            # a plugin can opt in to implicit tasks (such as meta). It does this
            # by declaring self.wants_implicit_tasks = True.
            if not callback_plugin.wants_implicit_tasks and (task_arg := self._first_arg_of_type(Task, args)) and task_arg.implicit:
                continue

            methods = []

            if method_name in callback_plugin._implemented_callback_methods:
                methods.append(getattr(callback_plugin, method_name))

            if 'v2_on_any' in callback_plugin._implemented_callback_methods:
                methods.append(getattr(callback_plugin, 'v2_on_any'))

            for method in methods:
                # send clean copies
                new_args = []

                for arg in args:
                    # FIXME: add play/task cleaners
                    if isinstance(arg, _RawTaskResult):
                        copied_tr = arg.as_callback_task_result()
                        new_args.append(copied_tr)
                        # this state hack requires that no callback ever accepts > 1 TaskResult object
                        callback_plugin._current_task_result = copied_tr
                    else:
                        new_args.append(arg)

                with self._callback_dispatch_error_handler.handle(AnsibleCallbackError):
                    try:
                        method(*new_args, **kwargs)
                    except AssertionError:
                        # Using an `assert` in integration tests is useful.
                        # Production code should never use `assert` or raise `AssertionError`.
                        raise
                    except Exception as ex:
                        raise AnsibleCallbackError(f"Callback dispatch {method_name!r} failed for plugin {callback_plugin._load_name!r}.") from ex

            callback_plugin._current_task_result = None  # clear temporary instance storage hack



"""Runtime patches for Jinja bugs affecting Ansible."""

from __future__ import annotations

import jinja2
import jinja2.utils


def _patch_jinja_undefined_slots() -> None:
    """
    Fix the broken __slots__ on Jinja's Undefined and StrictUndefined if they're missing in the current version.
    This will no longer be necessary once the fix is included in the minimum supported Jinja version.
    See: https://github.com/pallets/jinja/issues/2025
    """
    if not hasattr(jinja2.Undefined, '__slots__'):
        jinja2.Undefined.__slots__ = (
            "_undefined_hint",
            "_undefined_obj",
            "_undefined_name",
            "_undefined_exception",
        )

    if not hasattr(jinja2.StrictUndefined, '__slots__'):
        jinja2.StrictUndefined.__slots__ = ()


def _patch_jinja_missing_type() -> None:
    """
    Fix the `jinja2.utils.missing` type to support pickling while remaining a singleton.
    This will no longer be necessary once the fix is included in the minimum supported Jinja version.
    See: https://github.com/pallets/jinja/issues/2027
    """
    if getattr(jinja2.utils.missing, '__reduce__')() != 'missing':

        def __reduce__(*_args):
            return 'missing'

        type(jinja2.utils.missing).__reduce__ = __reduce__


def _patch_jinja() -> None:
    """Apply Jinja2 patches."""
    _patch_jinja_undefined_slots()
    _patch_jinja_missing_type()



from __future__ import annotations as _annotations

import typing as _t

from ansible.module_utils._internal._json import _profiles
from ansible._internal._json._profiles import _legacy
from ansible.parsing import vault as _vault


class LegacyControllerJSONEncoder(_legacy.Encoder):
    """Compatibility wrapper over `legacy` profile JSON encoder to support trust stripping and vault value plaintext conversion."""

    def __init__(self, preprocess_unsafe: bool = False, vault_to_text: bool = False, _decode_bytes: bool = False, **kwargs) -> None:
        self._preprocess_unsafe = preprocess_unsafe
        self._vault_to_text = vault_to_text
        self._decode_bytes = _decode_bytes

        super().__init__(**kwargs)

    def default(self, o: _t.Any) -> _t.Any:
        """Hooked default that can conditionally bypass base encoder behavior based on this instance's config."""
        if type(o) is _profiles._WrappedValue:  # pylint: disable=unidiomatic-typecheck
            o = o.wrapped

        if not self._preprocess_unsafe and type(o) is _legacy._Untrusted:  # pylint: disable=unidiomatic-typecheck
            return o.value  # if not emitting unsafe markers, bypass custom unsafe serialization and just return the raw value

        if self._vault_to_text and type(o) is _vault.EncryptedString:  # pylint: disable=unidiomatic-typecheck
            return str(o)  # decrypt and return the plaintext (or fail trying)

        if self._decode_bytes and isinstance(o, bytes):
            return o.decode(errors='surrogateescape')  # backward compatibility with `ansible.module_utils.basic.jsonify`

        return super().default(o)



# (c) 2012, Michael DeHaan <michael.dehaan@gmail.com>
# (c) 2013, Steven Dossett <sdossett@panath.com>
# (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

DOCUMENTATION = """
    name: inventory_hostnames
    author:
      - Michael DeHaan
      - Steven Dossett (!UNKNOWN) <sdossett@panath.com>
    version_added: "1.3"
    short_description: list of inventory hosts matching a host pattern
    description:
      - "This lookup understands 'host patterns' as used by the C(hosts:) keyword in plays
        and can return a list of matching hosts from inventory"
    notes:
      - this is only worth for 'hostname patterns' it is easier to loop over the group/group_names variables otherwise.
"""

EXAMPLES = """
- name: show all the hosts matching the pattern, i.e. all but the group www
  ansible.builtin.debug:
    msg: "{{ item }}"
  with_inventory_hostnames:
    - all:!www
"""

RETURN = """
 _hostnames:
    description: list of hostnames that matched the host pattern in inventory
    type: list
"""

from ansible.errors import AnsibleError
from ansible.inventory.manager import InventoryManager
from ansible.plugins.lookup import LookupBase


class LookupModule(LookupBase):
    def run(self, terms, variables=None, **kwargs):
        manager = InventoryManager(self._loader, parse=False)
        for group, hosts in variables['groups'].items():
            manager.add_group(group)
            for host in hosts:
                manager.add_host(host, group=group)

        try:
            return [h.name for h in manager.get_hosts(pattern=terms)]
        except AnsibleError:
            return []



from __future__ import annotations


class Broken:
    @property
    def accept_args_markers(self):
        raise Exception('boom')


class TestModule:
    def tests(self):
        return {
            'broken': Broken(),
        }



# (c) 2014 Michael DeHaan, <michael@ansible.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import os

from ansible import constants as C
from ansible.errors import AnsibleError, AnsibleAssertionError
from ansible.module_utils._internal._datatag import AnsibleTagHelper
from ansible.module_utils.six import string_types
from ansible.playbook.attribute import NonInheritableFieldAttribute
from ansible.playbook.base import Base
from ansible.playbook.collectionsearch import CollectionSearch
from ansible.playbook.conditional import Conditional
from ansible.playbook.taggable import Taggable
from ansible._internal._templating._engine import TemplateEngine
from ansible.utils.collection_loader import AnsibleCollectionRef
from ansible.utils.collection_loader._collection_finder import _get_collection_role_path
from ansible.utils.path import unfrackpath
from ansible.utils.display import Display

__all__ = ['RoleDefinition']

display = Display()


class RoleDefinition(Base, Conditional, Taggable, CollectionSearch):

    role = NonInheritableFieldAttribute(isa='string')

    def __init__(self, play=None, role_basedir=None, variable_manager=None, loader=None, collection_list=None):

        super(RoleDefinition, self).__init__()

        self._play = play
        self._variable_manager = variable_manager
        self._loader = loader

        self._role_path = None
        self._role_collection = None
        self._role_basedir = role_basedir
        self._role_params = dict()
        self._collection_list = collection_list

    # def __repr__(self):
    #     return 'ROLEDEF: ' + self._attributes.get('role', '<no name set>')

    @staticmethod
    def load(data, variable_manager=None, loader=None):
        raise AnsibleError("not implemented")

    def preprocess_data(self, ds):
        # role names that are simply numbers can be parsed by PyYAML
        # as integers even when quoted, so turn it into a string type
        if isinstance(ds, int):
            ds = "%s" % ds

        if not isinstance(ds, dict) and not isinstance(ds, string_types):
            raise AnsibleAssertionError()

        if isinstance(ds, dict):
            ds = super(RoleDefinition, self).preprocess_data(ds)

        # save the original ds for use later
        self._ds = ds

        # the new, cleaned datastructure, which will have legacy items reduced to a standard structure suitable for the
        # attributes of the task class; copy any tagged data to preserve things like origin
        new_ds = AnsibleTagHelper.tag_copy(ds, {})

        # first we pull the role name out of the data structure,
        # and then use that to determine the role path (which may
        # result in a new role name, if it was a file path)
        role_name = self._load_role_name(ds)
        (role_name, role_path) = self._load_role_path(role_name)

        # next, we split the role params out from the valid role
        # attributes and update the new datastructure with that
        # result and the role name
        if isinstance(ds, dict):
            (new_role_def, role_params) = self._split_role_params(ds)
            new_ds |= new_role_def
            self._role_params = role_params

        # set the role name in the new ds
        new_ds['role'] = role_name

        # we store the role path internally
        self._role_path = role_path

        # and return the cleaned-up data structure
        return new_ds

    def _load_role_name(self, ds):
        """
        Returns the role name (either the role: or name: field) from
        the role definition, or (when the role definition is a simple
        string), just that string
        """

        if isinstance(ds, string_types):
            return ds

        role_name = ds.get('role', ds.get('name'))
        if not role_name or not isinstance(role_name, string_types):
            raise AnsibleError('role definitions must contain a role name', obj=ds)

        # if we have the required datastructures, and if the role_name
        # contains a variable, try and template it now
        if self._variable_manager:
            all_vars = self._variable_manager.get_vars(play=self._play)
            templar = TemplateEngine(loader=self._loader, variables=all_vars)
            role_name = templar.template(role_name)

        return role_name

    def _load_role_path(self, role_name):
        """
        the 'role', as specified in the ds (or as a bare string), can either
        be a simple name or a full path. If it is a full path, we use the
        basename as the role name, otherwise we take the name as-given and
        append it to the default role path
        """

        # create a templar class to template the dependency names, in
        # case they contain variables
        if self._variable_manager is not None:
            all_vars = self._variable_manager.get_vars(play=self._play)
        else:
            all_vars = dict()

        templar = TemplateEngine(loader=self._loader, variables=all_vars)
        role_name = templar.template(role_name)

        role_tuple = None

        # try to load as a collection-based role first
        if self._collection_list or AnsibleCollectionRef.is_valid_fqcr(role_name):
            role_tuple = _get_collection_role_path(role_name, self._collection_list)

        if role_tuple:
            # we found it, stash collection data and return the name/path tuple
            self._role_collection = role_tuple[2]
            return role_tuple[0:2]

        # We didn't find a collection role, look in defined role paths
        # FUTURE: refactor this to be callable from internal so we can properly order
        # ansible.legacy searches with the collections keyword

        # we always start the search for roles in the base directory of the playbook
        role_search_paths = [
            os.path.join(self._loader.get_basedir(), u'roles'),
        ]

        # also search in the configured roles path
        if C.DEFAULT_ROLES_PATH:
            role_search_paths.extend(C.DEFAULT_ROLES_PATH)

        # next, append the roles basedir, if it was set, so we can
        # search relative to that directory for dependent roles
        if self._role_basedir:
            role_search_paths.append(self._role_basedir)

        # finally as a last resort we look in the current basedir as set
        # in the loader (which should be the playbook dir itself) but without
        # the roles/ dir appended
        role_search_paths.append(self._loader.get_basedir())

        # now iterate through the possible paths and return the first one we find
        for path in role_search_paths:
            path = templar.template(path)
            role_path = unfrackpath(os.path.join(path, role_name))
            if self._loader.path_exists(role_path):
                return (role_name, role_path)

        # if not found elsewhere try to extract path from name
        role_path = unfrackpath(role_name)
        if self._loader.path_exists(role_path):
            role_name = os.path.basename(role_name)
            return (role_name, role_path)

        searches = (self._collection_list or []) + role_search_paths

        raise AnsibleError("the role '%s' was not found in %s" % (role_name, ":".join(searches)), obj=self._ds)

    def _split_role_params(self, ds):
        """
        Splits any random role params off from the role spec and store
        them in a dictionary of params for parsing later
        """

        role_def = dict()
        role_params = dict()
        base_attribute_names = frozenset(self.fattributes)
        for (key, value) in ds.items():
            # use the list of FieldAttribute values to determine what is and is not
            # an extra parameter for this role (or sub-class of this role)
            # FIXME: hard-coded list of exception key names here corresponds to the
            #        connection fields in the Base class. There may need to be some
            #        other mechanism where we exclude certain kinds of field attributes,
            #        or make this list more automatic in some way so we don't have to
            #        remember to update it manually.
            if key not in base_attribute_names:
                # this key does not match a field attribute, so it must be a role param
                role_params[key] = value
            else:
                # this is a field attribute, so copy it over directly
                role_def[key] = value

        return (role_def, role_params)

    def get_role_params(self):
        return self._role_params.copy()

    def get_role_path(self):
        return self._role_path

    def get_name(self, include_role_fqcn=True):
        if include_role_fqcn:
            return '.'.join(x for x in (self._role_collection, self.role) if x)
        return self.role



from __future__ import annotations

from ansible._internal._datatag import _tags
from ansible.module_utils.common.text.converters import to_bytes
from ansible.parsing.vault import VaultSecret, VaultSecretsContext, VaultLib, EncryptedString


class TextVaultSecret(VaultSecret):
    """A secret piece of text. ie, a password. Tracks text encoding.

    The text encoding of the text may not be the default text encoding so
    we keep track of the encoding so we encode it to the same bytes."""

    def __init__(self, text, encoding=None, errors=None, _bytes=None):
        super(TextVaultSecret, self).__init__()
        self.text = text
        self.encoding = encoding or 'utf-8'
        self._bytes = _bytes
        self.errors = errors or 'strict'

    @property
    def bytes(self):
        """The text encoded with encoding, unless we specifically set _bytes."""
        return self._bytes or to_bytes(self.text, encoding=self.encoding, errors=self.errors)


class VaultTestHelper:
    @classmethod
    def make_vault_ciphertext(cls, plaintext: str) -> str:
        """Creates an `EncryptedString` from the first secret in the active VaultSecretsContext."""
        secrets = VaultSecretsContext.current().secrets
        vl = VaultLib(secrets)

        return vl.encrypt(plaintext, secrets[0][1]).decode()

    @classmethod
    def make_encrypted_string(cls, plaintext: str) -> EncryptedString:
        """Creates an `EncryptedString` from the first secret in the active VaultSecretsContext."""
        return _tags.Origin(path="/tmp/sometest", line_num=42, col_num=42).tag(EncryptedString(ciphertext=cls.make_vault_ciphertext(plaintext)))



from __future__ import annotations

import contextlib as _contextlib
import dataclasses
import typing as t

from ansible.module_utils._internal._datatag import AnsibleSingletonTagBase, _tag_dataclass_kwargs
from ansible.module_utils._internal._datatag._tags import Deprecated
from ansible._internal._datatag._tags import Origin
from ansible.utils.display import Display

from ._access import NotifiableAccessContextBase
from ._utils import TemplateContext

display = Display()


@dataclasses.dataclass(**_tag_dataclass_kwargs)
class _JinjaConstTemplate(AnsibleSingletonTagBase):
    # deprecated: description='embedded Jinja constant string template support' core_version='2.23'
    pass


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class _TrippedDeprecationInfo:
    template: str
    deprecated: Deprecated


class DeprecatedAccessAuditContext(NotifiableAccessContextBase):
    """When active, captures metadata about managed accesses to `Deprecated` tagged objects."""

    _type_interest = frozenset([Deprecated])

    @classmethod
    def when(cls, condition: bool, /) -> t.Self | _contextlib.nullcontext:
        """Returns a new instance if `condition` is True (usually `TemplateContext.is_top_level`), otherwise a `nullcontext` instance."""
        if condition:
            return cls()

        return _contextlib.nullcontext()

    def __init__(self) -> None:
        self._tripped_deprecation_info: dict[int, _TrippedDeprecationInfo] = {}

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        result = super().__exit__(exc_type, exc_val, exc_tb)

        for item in self._tripped_deprecation_info.values():
            if Origin.is_tagged_on(item.template):
                msg = item.deprecated.msg
            else:
                # without an origin, we need to include what context we do have (the template)
                msg = f'While processing {item.template!r}: {item.deprecated.msg}'

            display._deprecated_with_plugin_info(
                msg=msg,
                help_text=item.deprecated.help_text,
                version=item.deprecated.version,
                date=item.deprecated.date,
                obj=item.template,
                deprecator=item.deprecated.deprecator,
                formatted_traceback=item.deprecated.formatted_traceback,
            )

        return result

    def _notify(self, o: t.Any) -> None:
        deprecated = Deprecated.get_required_tag(o)
        deprecated_key = id(deprecated)

        if deprecated_key in self._tripped_deprecation_info:
            return  # record only the first access for each deprecated tag in a given context

        template_ctx = TemplateContext.current(optional=True)
        template = template_ctx.template_value if template_ctx else None

        # when the current template input is a container, provide a descriptive string with origin propagated (if possible)
        if not isinstance(template, str):
            # DTFIX-FUTURE: ascend the template stack to try and find the nearest string source template
            origin = Origin.get_tag(template)

            # DTFIX-FUTURE: this should probably use a synthesized description value on the tag
            #              it is reachable from the data_tagging_controller test: ../playbook_output_validator/filter.py actual_stdout.txt actual_stderr.txt
            # -[DEPRECATION WARNING]: `something_old` is deprecated, don't use it! This feature will be removed in version 1.2.3.
            # +[DEPRECATION WARNING]: While processing '<<container>>': `something_old` is deprecated, don't use it! This feature will be removed in ...
            template = '<<container>>'

            if origin:
                origin.tag(template)

        self._tripped_deprecation_info[deprecated_key] = _TrippedDeprecationInfo(
            template=template,
            deprecated=deprecated,
        )



"""Proxy stdlib threading module that only supports non-joinable daemon threads."""

# NB: all new local module attrs are _ prefixed to ensure an identical public attribute surface area to the module we're proxying

from __future__ import annotations as _annotations

import threading as _threading
import typing as _t


class _DaemonThread(_threading.Thread):
    """
    Daemon-only Thread subclass; prevents running threads of this type from blocking interpreter shutdown and process exit.
    The join() method is a no-op.
    """

    def __init__(self, *args, daemon: bool | None = None, **kwargs) -> None:
        super().__init__(*args, daemon=daemon or True, **kwargs)

    def join(self, timeout=None) -> None:
        """ThreadPoolExecutor's atexit handler joins all queue threads before allowing shutdown; prevent them from blocking."""


Thread = _DaemonThread  # shadow the real Thread attr with our _DaemonThread


def __getattr__(name: str) -> _t.Any:
    """Delegate anything not defined locally to the real `threading` module."""
    return getattr(_threading, name)



from __future__ import annotations

import datetime as _datetime

from ansible.module_utils._internal import _datatag
from ansible.module_utils._internal._json import _profiles
from ansible.parsing import vault as _vault
from ansible._internal._datatag import _tags


class _Profile(_profiles._JSONSerializationProfile):
    """Profile for external cache persistence of inventory/fact data that preserves most tags."""

    serialize_map = {}
    schema_id = 1

    @classmethod
    def post_init(cls, **kwargs):
        cls.allowed_ansible_serializable_types = (
            _profiles._common_module_types
            | _profiles._common_module_response_types
            | {
                _datatag._AnsibleTaggedDate,
                _datatag._AnsibleTaggedTime,
                _datatag._AnsibleTaggedDateTime,
                _datatag._AnsibleTaggedStr,
                _datatag._AnsibleTaggedInt,
                _datatag._AnsibleTaggedFloat,
                _datatag._AnsibleTaggedList,
                _datatag._AnsibleTaggedSet,
                _datatag._AnsibleTaggedTuple,
                _datatag._AnsibleTaggedDict,
                _tags.SourceWasEncrypted,
                _tags.Origin,
                _tags.TrustedAsTemplate,
                _vault.EncryptedString,
                _vault.VaultedValue,
            }
        )

        cls.serialize_map = {
            set: cls.serialize_as_list,
            tuple: cls.serialize_as_list,
            _datetime.date: _datatag.AnsibleSerializableDate,
            _datetime.time: _datatag.AnsibleSerializableTime,
            _datetime.datetime: _datatag.AnsibleSerializableDateTime,
        }

        cls.handle_key = cls._handle_key_str_fallback  # legacy stdlib-compatible key behavior


class Encoder(_profiles.AnsibleProfileJSONEncoder):
    _profile = _Profile


class Decoder(_profiles.AnsibleProfileJSONDecoder):
    _profile = _Profile



# Copyright: (c) Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import annotations

import unittest

from ansible.errors import AnsibleError
from ansible.playbook.task import Task


basic_command_task = dict(
    name='Test Task',
    command='echo hi'
)

kv_command_task = dict(
    action='command echo hi'
)


class TestTask(unittest.TestCase):
    def test_load_task_simple(self):
        t = Task.load(basic_command_task)
        assert t is not None
        self.assertEqual(t.get_name(), basic_command_task['name'])
        self.assertEqual(t.action, 'command')
        self.assertEqual(t.args, dict(_raw_params='echo hi'))

    def test_load_task_kv_form(self):
        t = Task.load(kv_command_task)
        self.assertEqual(t.action, 'command')
        self.assertEqual(t.args, dict(_raw_params='echo hi'))

    def test_task_auto_name(self):
        assert 'name' not in kv_command_task
        Task.load(kv_command_task)
        self.assertNotIn('name', kv_command_task)
        t = Task.load(kv_command_task)
        self.assertEqual(t.get_name(), 'command')

    def test_delay(self):
        task_base = {'name': 'test', 'action': 'debug'}
        good_params = [
            (0, 0),
            (0.1, 0.1),
            ('0.3', 0.3),
            ('0.03', 0.03),
            ('12', 12),
            (12, 12),
            (1.2, 1.2),
            ('1.2', 1.2),
            ('1.0', 1),
        ]
        for delay, expected in good_params:
            with self.subTest(f'type "{type(delay)}" was not cast to float', delay=delay, expected=expected):
                p = dict(delay=delay)
                p.update(task_base)
                t = Task().load_data(p)
                self.assertEqual(t.get_validated_value('delay', t.fattributes.get('delay'), delay, None), expected)

        bad_params = [
            'E',
            '1.E',
            'E.1',
        ]
        for delay in bad_params:
            with self.subTest(f'type "{type(delay)} was cast to float w/o error', delay=delay):
                p = dict(delay=delay)
                p.update(task_base)
                t = Task().load_data(p)
                with self.assertRaises(AnsibleError):
                    dummy = t.get_validated_value('delay', t.fattributes.get('delay'), delay, None)

    def test_task_auto_name_with_role(self):
        pass

    def test_load_task_complex_form(self):
        pass

    def test_can_load_module_complex_form(self):
        pass

    def test_local_action_implies_delegate(self):
        pass

    def test_local_action_conflicts_with_delegate(self):
        pass

    def test_delegate_to_parses(self):
        pass



# (c) 2012-2014, Michael DeHaan <michael.dehaan@gmail.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import collections.abc as c
import typing as t

from ansible.module_utils._internal import _datatag

if t.TYPE_CHECKING:
    from ansible.inventory.manager import InventoryManager
    from ansible.parsing.dataloader import DataLoader
    from ansible._internal._templating._jinja_common import Marker
    from ansible.vars.manager import VariableManager


__all__ = ['HostVars', 'HostVarsVars']


class HostVars(c.Mapping):
    """A read-only wrapper to enable on-demand templating of a specific host's variables under that host's variable context."""
    def __init__(self, inventory: InventoryManager, variable_manager: VariableManager, loader: DataLoader) -> None:
        self._inventory = inventory
        self._loader = loader
        self._variable_manager = variable_manager

        variable_manager._hostvars = self

    def raw_get(self, host_name: str) -> dict[str, t.Any] | Marker:
        """
        Similar to __getitem__, however the returned data is not run through
        the templating engine to expand variables in the hostvars.
        """
        # does not use inventory.hosts, so it can create localhost on demand
        host = self._inventory.get_host(host_name)

        if host is None:
            from ansible._internal._templating import _jinja_bits

            return _jinja_bits._undef(f"hostvars[{host_name!r}]")

        return self._variable_manager.get_vars(host=host, include_hostvars=False)

    def __getitem__(self, key: str) -> HostVarsVars | Marker:
        data = self.raw_get(key)

        from ansible._internal._templating import _jinja_bits

        if isinstance(data, _jinja_bits.Marker):
            return data

        return HostVarsVars(data, loader=self._loader, host=key)

    def __contains__(self, item: object) -> bool:
        # does not use inventory.hosts, so it can create localhost on demand
        return self._inventory.get_host(item) is not None

    def __iter__(self) -> t.Iterator[str]:
        yield from self._inventory.hosts

    def __len__(self) -> int:
        return len(self._inventory.hosts)

    def __deepcopy__(self, memo: t.Any) -> HostVars:
        # this object may be stored in a var dict that is itself deep copied, but since the underlying data
        # is supposed to be immutable, we don't need to actually copy the data
        return self


class HostVarsVars(c.Mapping):
    """A read-only view of a specific host's vars that will template on access under that host's variable context."""

    def __init__(self, variables: dict[str, t.Any], loader: DataLoader | None, host: str) -> None:
        from ansible._internal._templating import _engine

        self._vars = variables
        self._templar = _engine.TemplateEngine(variables=variables, loader=loader)
        self._host = host

    def __getitem__(self, key: str) -> t.Any:
        return self._templar.template(self._vars[key])

    def __contains__(self, item: object) -> bool:
        return item in self._vars

    def __iter__(self) -> t.Iterator[str]:
        yield from self._vars

    def __len__(self) -> int:
        return len(self._vars)

    def __repr__(self) -> str:
        return f'HostVarsVars({self._host=!r}, ...)'

    def __deepcopy__(self, memo: t.Any) -> HostVarsVars:
        # this may be stored in a var dict that is itself deep copied, but since the underlying data
        # is supposed to be immutable, we don't need to actually copy the data
        return self


# DTFIX-FUTURE: is there a better way to add this to the ignorable types in the module_utils code
_datatag._untaggable_types.update({HostVars, HostVarsVars})



"""Command line parsing for the `coverage analyze targets filter` command."""

from __future__ import annotations

import argparse

from ......commands.coverage.analyze.targets.filter import (
    command_coverage_analyze_targets_filter,
    CoverageAnalyzeTargetsFilterConfig,
)

from .....environments import (
    CompositeActionCompletionFinder,
    ControllerMode,
    TargetMode,
    add_environments,
)


def do_filter(
    subparsers,
    parent: argparse.ArgumentParser,
    completer: CompositeActionCompletionFinder,
):
    """Command line parsing for the `coverage analyze targets filter` command."""
    parser: argparse.ArgumentParser = subparsers.add_parser(
        'filter',
        parents=[parent],
        help='filter aggregated coverage data',
    )

    parser.set_defaults(
        func=command_coverage_analyze_targets_filter,
        config=CoverageAnalyzeTargetsFilterConfig,
    )

    targets_filter = parser.add_argument_group(title='coverage arguments')

    targets_filter.add_argument(
        'input_file',
        help='input file to read aggregated coverage from',
    )

    targets_filter.add_argument(
        'output_file',
        help='output file to write expanded coverage to',
    )

    targets_filter.add_argument(
        '--include-target',
        metavar='TGT',
        dest='include_targets',
        action='append',
        help='include the specified targets',
    )

    targets_filter.add_argument(
        '--exclude-target',
        metavar='TGT',
        dest='exclude_targets',
        action='append',
        help='exclude the specified targets',
    )

    targets_filter.add_argument(
        '--include-path',
        metavar='REGEX',
        help='include paths matching the given regex',
    )

    targets_filter.add_argument(
        '--exclude-path',
        metavar='REGEX',
        help='exclude paths matching the given regex',
    )

    add_environments(parser, completer, ControllerMode.ORIGIN, TargetMode.NO_TARGETS)  # coverage analyze targets filter



# (c) 2015, Marius Gedminas <marius@gedmin.as>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import pytest

from datetime import datetime

from ansible.utils.helpers import pct_to_int, object_to_dict, deduplicate_list


pct_to_int_testdata = [
    pytest.param(
        1, 100, 1, 1, id="positive_percentage"
    ),
    pytest.param(
        -1, 100, 1, -1, id="negative_percentage"
    ),
    pytest.param(
        "1%", 10, 1, 1, id="string_percentage"
    ),
    pytest.param(
        "1%", 10, 0, 0, id="string_percentage_with_zero_min_value"
    ),
    pytest.param(
        "1", 100, 1, 1, id="string_percentage_without_sign"
    ),
    pytest.param(
        "10%", 100, 1, 10, id="string_percentage_two_digit"
    )
]


@pytest.mark.parametrize("value,num_items,min_value,expected", pct_to_int_testdata)
def test_pct_to_int(value, num_items, min_value, expected):
    assert pct_to_int(value, num_items, min_value) == expected


def test_object_to_dict():
    test_dict = object_to_dict(datetime(2024, 7, 30))
    assert test_dict['day'] == 30
    assert test_dict['year'] == 2024
    assert test_dict['month'] == 7

    test_dict_without_day = object_to_dict(datetime(2024, 7, 30), exclude=['day'])
    assert 'day' not in list(test_dict_without_day.keys())


def test_deduplicate_list():
    assert deduplicate_list([1, 2, 2, 3]) == [1, 2, 3]
    assert deduplicate_list([1, 2, 3]) == [1, 2, 3]



from __future__ import annotations

import pytest

from ansible.errors import AnsibleTemplateError
from ansible.template import Templar, trust_as_template


@pytest.mark.parametrize("value", (
    "not_defined is not defined",  # direct rendered undefined template
    "not_defined is undefined",
    "chained_undefined is not defined",  # chain rendered undefined template
    "chained_undefined is undefined",
    "valid_scalar is defined",  # valid scalar
    "valid_scalar is not undefined",
    "valid_template is defined",  # valid chain rendered template
    "valid_template is not undefined",  # valid chain rendered template
))
def test_defined_undefined_success(value):
    """Validate success behavior for the `defined` and `undefined` Jinja test implementations."""
    variables = dict(
        valid_scalar="valid",
        valid_template=trust_as_template("{{ 'hey' }}"),
        chained_undefined=trust_as_template("{{ bogus }}"),
    )

    assert Templar(variables=variables).evaluate_conditional(trust_as_template(value))


@pytest.mark.parametrize("value", (
    "syntax_error is defined",
    "syntax_error is undefined",
    "div_by_zero is defined",
    "div_by_zero is undefined",
))
def test_defined_undefined_failure(value):
    variables = dict(
        syntax_error=trust_as_template("{{ / }}"),
        div_by_zero=trust_as_template("{{ 1 / 0 }}"),
    )

    with pytest.raises(AnsibleTemplateError):
        Templar(variables=variables).evaluate_conditional(trust_as_template(value))



from __future__ import annotations


def do_flag(myval):
    return 'flagged'


class FilterModule(object):
    """ Ansible core jinja2 filters """

    def filters(self):
        return {
            # jinja2 overrides
            'flag': do_flag,
            'flatten': do_flag,
        }



# (c) 2012, Michael DeHaan <michael.dehaan@gmail.com>
# (c) 2013, Dylan Martin <dmartin@seattlecentral.edu>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import os

from ansible.errors import AnsibleActionFail, AnsibleActionSkip
from ansible.module_utils.parsing.convert_bool import boolean
from ansible.plugins.action import ActionBase


class ActionModule(ActionBase):

    TRANSFERS_FILES = True

    def run(self, tmp=None, task_vars=None):
        """ handler for unarchive operations """
        if task_vars is None:
            task_vars = dict()

        super(ActionModule, self).run(tmp, task_vars)
        del tmp  # tmp no longer has any effect

        source = self._task.args.get('src', None)
        dest = self._task.args.get('dest', None)
        remote_src = boolean(self._task.args.get('remote_src', False), strict=False)
        creates = self._task.args.get('creates', None)
        decrypt = self._task.args.get('decrypt', True)

        try:
            # "copy" is deprecated in favor of "remote_src".
            if 'copy' in self._task.args:
                # They are mutually exclusive.
                if 'remote_src' in self._task.args:
                    raise AnsibleActionFail("parameters are mutually exclusive: ('copy', 'remote_src')")
                # We will take the information from copy and store it in
                # the remote_src var to use later in this file.
                self._task.args['remote_src'] = remote_src = not boolean(self._task.args.pop('copy'), strict=False)

            if source is None or dest is None:
                raise AnsibleActionFail("src (or content) and dest are required")

            if creates:
                # do not run the command if the line contains creates=filename
                # and the filename already exists. This allows idempotence
                # of command executions.
                creates = self._remote_expand_user(creates)
                if self._remote_file_exists(creates):
                    raise AnsibleActionSkip("skipped, since %s exists" % creates)

            dest = self._remote_expand_user(dest)  # CCTODO: Fix path for Windows hosts.
            source = os.path.expanduser(source)

            if not remote_src:
                source = self._loader.get_real_file(self._find_needle('files', source), decrypt=decrypt)

            remote_stat = self._execute_remote_stat(dest, all_vars=task_vars, follow=True)

            if not remote_stat['exists'] or not remote_stat['isdir']:
                raise AnsibleActionFail("dest '%s' must be an existing dir" % dest)

            if not remote_src:
                # transfer the file to a remote tmp location
                tmp_src = self._connection._shell.join_path(self._connection._shell.tmpdir, 'source')
                self._transfer_file(source, tmp_src)

            # handle diff mode client side
            # handle check mode client side

            # remove action plugin only keys
            new_module_args = self._task.args.copy()
            for key in ('decrypt',):
                if key in new_module_args:
                    del new_module_args[key]

            if not remote_src:
                # fix file permissions when the copy is done as a different user
                self._fixup_perms2((self._connection._shell.tmpdir, tmp_src))
                new_module_args['src'] = tmp_src

            # execute the unarchive module now, with the updated args (using ansible.legacy prefix to eliminate collections
            # collisions with local override
            return self._execute_module(module_name='ansible.legacy.unarchive', module_args=new_module_args, task_vars=task_vars)
        finally:
            self._remove_tmp_path(self._connection._shell.tmpdir)



# -*- coding: utf-8 -*-
# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import pytest

from ansible.module_utils.facts.network import generic_bsd


def mock_get_bin_path(command):
    cmds = {
        'ifconfig': 'fake/ifconfig',
        'route': 'fake/route',
    }
    return cmds.get(command, None)


NETBSD_IFCONFIG_A_OUT_7_1 = r"""
lo0: flags=8049<UP,LOOPBACK,RUNNING,MULTICAST> mtu 33624
        inet 127.0.0.1 netmask 0xff000000
        inet6 ::1 prefixlen 128
        inet6 fe80::1%lo0 prefixlen 64 scopeid 0x1
re0: flags=8843<UP,BROADCAST,RUNNING,SIMPLEX,MULTICAST> mtu 1500
        capabilities=3f80<TSO4,IP4CSUM_Rx,IP4CSUM_Tx,TCP4CSUM_Rx,TCP4CSUM_Tx>
        capabilities=3f80<UDP4CSUM_Rx,UDP4CSUM_Tx>
        enabled=0
        ec_capabilities=3<VLAN_MTU,VLAN_HWTAGGING>
        ec_enabled=0
        address: 52:54:00:63:55:af
        media: Ethernet autoselect (100baseTX full-duplex)
        status: active
        inet 192.168.122.205 netmask 0xffffff00 broadcast 192.168.122.255
        inet6 fe80::5054:ff:fe63:55af%re0 prefixlen 64 scopeid 0x2
"""

NETBSD_IFCONFIG_A_OUT_POST_7_1 = r"""
lo0: flags=0x8049<UP,LOOPBACK,RUNNING,MULTICAST> mtu 33624
        inet 127.0.0.1/8 flags 0x0
        inet6 ::1/128 flags 0x20<NODAD>
        inet6 fe80::1%lo0/64 flags 0x0 scopeid 0x1
re0: flags=0x8843<UP,BROADCAST,RUNNING,SIMPLEX,MULTICAST> mtu 1500
        capabilities=3f80<TSO4,IP4CSUM_Rx,IP4CSUM_Tx,TCP4CSUM_Rx,TCP4CSUM_Tx>
        capabilities=3f80<UDP4CSUM_Rx,UDP4CSUM_Tx>
        enabled=0
        ec_capabilities=3<VLAN_MTU,VLAN_HWTAGGING>
        ec_enabled=0
        address: 52:54:00:63:55:af
        media: Ethernet autoselect (100baseTX full-duplex)
        status: active
        inet 192.168.122.205/24 broadcast 192.168.122.255 flags 0x0
        inet6 fe80::5054:ff:fe63:55af%re0/64 flags 0x0 scopeid 0x2
"""

NETBSD_EXPECTED = {'all_ipv4_addresses': ['192.168.122.205'],
                   'all_ipv6_addresses': ['fe80::5054:ff:fe63:55af%re0'],
                   'default_ipv4': {},
                   'default_ipv6': {},
                   'interfaces': ['lo0', 're0'],
                   'lo0': {'device': 'lo0',
                           'flags': ['UP', 'LOOPBACK', 'RUNNING', 'MULTICAST'],
                           'ipv4': [{'address': '127.0.0.1',
                                     'broadcast': '127.255.255.255',
                                     'netmask': '255.0.0.0',
                                     'network': '127.0.0.0'}],
                           'ipv6': [{'address': '::1', 'prefix': '128'},
                                    {'address': 'fe80::1%lo0', 'prefix': '64', 'scope': '0x1'}],
                           'macaddress': 'unknown',
                           'mtu': '33624',
                           'type': 'loopback'},
                   're0': {'device': 're0',
                           'flags': ['UP', 'BROADCAST', 'RUNNING', 'SIMPLEX', 'MULTICAST'],
                           'ipv4': [{'address': '192.168.122.205',
                                     'broadcast': '192.168.122.255',
                                     'netmask': '255.255.255.0',
                                     'network': '192.168.122.0'}],
                           'ipv6': [{'address': 'fe80::5054:ff:fe63:55af%re0',
                                     'prefix': '64',
                                     'scope': '0x2'}],
                           'macaddress': 'unknown',
                           'media': 'Ethernet',
                           'media_options': [],
                           'media_select': 'autoselect',
                           'media_type': '100baseTX',
                           'mtu': '1500',
                           'status': 'active',
                           'type': 'ether'}}


@pytest.mark.parametrize(
    ("test_input", "expected"),
    [
        pytest.param(
            {
                "mock_run_command": [
                    (0, "Foo", ""),
                    (0, "Foo", ""),
                    (0, NETBSD_IFCONFIG_A_OUT_7_1, ""),
                ],
            },
            NETBSD_EXPECTED,
            id="old-ifconfig",
        ),
        pytest.param(
            {
                "mock_run_command": [
                    (0, "Foo", ""),
                    (0, "Foo", ""),
                    (0, NETBSD_IFCONFIG_A_OUT_POST_7_1, ""),
                ],
            },
            NETBSD_EXPECTED,
            id="post-7-1-ifconfig",
        ),
    ],
)
def test_generic_bsd_ifconfig(mocker, test_input, expected):
    module = mocker.MagicMock()
    mocker.patch.object(module, "get_bin_path", side_effect=mock_get_bin_path)
    mocker.patch.object(
        module, "run_command", side_effect=test_input["mock_run_command"]
    )

    bsd_net = generic_bsd.GenericBsdIfconfigNetwork(module)
    res = bsd_net.populate()
    assert res == expected


def test_compare_old_new_ifconfig(mocker):
    old_ifconfig_module = mocker.MagicMock()
    mocker.patch.object(old_ifconfig_module, "get_bin_path", side_effect=mock_get_bin_path)
    mocker.patch.object(
        old_ifconfig_module,
        "run_command",
        side_effect=[
            (0, "Foo", ""),
            (0, "Foo", ""),
            (0, NETBSD_IFCONFIG_A_OUT_7_1, ""),
        ],
    )
    old_bsd_net = generic_bsd.GenericBsdIfconfigNetwork(old_ifconfig_module)
    old_res = old_bsd_net.populate()

    new_ifconfig_module = mocker.MagicMock()
    mocker.patch.object(new_ifconfig_module, "get_bin_path", side_effect=mock_get_bin_path)
    mocker.patch.object(
        new_ifconfig_module,
        "run_command",
        side_effect=[
            (0, "Foo", ""),
            (0, "Foo", ""),
            (0, NETBSD_IFCONFIG_A_OUT_POST_7_1, ""),
        ],
    )
    new_bsd_net = generic_bsd.GenericBsdIfconfigNetwork(new_ifconfig_module)
    new_res = new_bsd_net.populate()
    assert old_res == new_res


@pytest.mark.parametrize(
    ("test_input", "expected"),
    [
        pytest.param(
            "inet 192.168.7.113 netmask 0xffffff00 broadcast 192.168.7.255",
            (
                {
                    'ipv4': [
                        {
                            'address': '192.168.7.113',
                            'netmask': '255.255.255.0',
                            'network': '192.168.7.0',
                            'broadcast': '192.168.7.255',
                        }
                    ]
                },
                {'all_ipv4_addresses': ['192.168.7.113']},
            ),
            id="ifconfig-output-1",
        ),
        pytest.param(
            "inet 10.109.188.206 --> 10.109.188.206 netmask 0xffffe000",
            (
                {
                    'ipv4': [
                        {
                            'address': '10.109.188.206',
                            'netmask': '255.255.224.0',
                            'network': '10.109.160.0',
                            'broadcast': '10.109.191.255',
                        }
                    ]
                },
                {'all_ipv4_addresses': ['10.109.188.206']},
            ),
            id="ifconfig-output-2",
        ),
    ],
)
def test_ensure_correct_netmask_parsing(test_input, expected):
    n = generic_bsd.GenericBsdIfconfigNetwork(None)
    words = test_input.split()
    current_if = {"ipv4": []}
    ips = {"all_ipv4_addresses": []}
    n.parse_inet_line(words, current_if, ips)
    assert current_if == expected[0]
    assert ips == expected[1]



# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# Determine if a system is in 'fips' mode

from __future__ import annotations

import typing as t

from ansible.module_utils.facts.utils import get_file_content

from ansible.module_utils.facts.collector import BaseFactCollector


class FipsFactCollector(BaseFactCollector):
    name = 'fips'
    _fact_ids = set()  # type: t.Set[str]

    def collect(self, module=None, collected_facts=None):
        # NOTE: this is populated even if it is not set
        fips_facts = {
            'fips': False
        }
        if get_file_content('/proc/sys/crypto/fips_enabled') == '1':
            fips_facts['fips'] = True
        return fips_facts



#!/usr/bin/python
from __future__ import annotations


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['removed'],
                    'supported_by': 'core'}

DOCUMENTATION = """
---
module: test_docs_removed_status
short_description: Test module
description:
    - Test module
author:
    - Ansible Core Team
"""

EXAMPLES = """
"""

RETURN = """
"""


from ansible.module_utils.basic import AnsibleModule


def main():
    module = AnsibleModule(
        argument_spec=dict(),
    )

    module.exit_json()


if __name__ == '__main__':
    main()



# (c) 2020 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

DOCUMENTATION = """
    name: display_resolved_action
    type: aggregate
    short_description: Displays the requested and resolved actions at the end of a playbook.
    description:
        - Displays the requested and resolved actions in the format "requested == resolved".
    options:
      test_on_task_start:
        description: Test using task.resolved_action before it is reliably resolved.
        default: False
        env:
          - name: ANSIBLE_TEST_ON_TASK_START
    requirements:
      - Enable in configuration.
"""

from ansible.plugins.callback import CallbackBase


class CallbackModule(CallbackBase):

    CALLBACK_VERSION = 2.0
    CALLBACK_TYPE = 'aggregate'
    CALLBACK_NAME = 'display_resolved_action'
    CALLBACK_NEEDS_ENABLED = True

    def __init__(self, *args, **kwargs):
        super(CallbackModule, self).__init__(*args, **kwargs)

    def v2_playbook_on_task_start(self, task, is_conditional):
        if self.get_option("test_on_task_start"):
            self._display.display(f"v2_playbook_on_task_start: {task.action} == {task.resolved_action}")

    def v2_runner_item_on_ok(self, result):
        self._display.display(f"v2_runner_item_on_ok: {result.task.action} == {result.task.resolved_action}")

    def v2_runner_on_ok(self, result):
        if not result.task.loop:
            self._display.display(f"v2_runner_on_ok: {result.task.action} == {result.task.resolved_action}")
